@@ -26,10 +26,13 @@ THE SOFTWARE.
 """
 
 
+from functools import reduce
 import operator
 import sys
 import logging
-from functools import reduce
+from typing import (
+        Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, TypeVar)
+
 
 import six
 from six.moves import range, zip, intern, input
@@ -63,8 +66,9 @@ Memoization
 .. autofunction:: memoize
 .. autofunction:: memoize_on_first_arg
 .. autofunction:: memoize_method
-.. autofunction:: memoize_method_with_uncached
 .. autofunction:: memoize_in
+.. autofunction:: keyed_memoize_on_first_arg
+.. autofunction:: keyed_memoize_method
 
 Argmin/max
 ----------
@@ -83,7 +87,6 @@ Permutations, Tuples, Integer sequences
 ---------------------------------------
 
 .. autofunction:: wandering_element
-.. autofunction:: indices_in_shape
 .. autofunction:: generate_nonnegative_integer_tuples_below
 .. autofunction:: generate_nonnegative_integer_tuples_summing_to_at_most
 .. autofunction:: generate_all_nonnegative_integer_tuples
@@ -152,6 +155,13 @@ Sorting in natural order
 .. autofunction:: natsorted
 """
 
+# {{{ type variables
+
+T = TypeVar("T")
+F = TypeVar('F', bound=Callable[..., Any])
+
+# }}}
+
 
 # {{{ math --------------------------------------------------------------------
 
@@ -236,7 +246,7 @@ class RecordWithoutPickling(object):
     will be individually derived from this class.
     """
 
-    __slots__ = []
+    __slots__: List[str] = []
 
     def __init__(self, valuedict=None, exclude=None, **kwargs):
         assert self.__class__ is not Record
@@ -293,7 +303,7 @@ class RecordWithoutPickling(object):
 
 
 class Record(RecordWithoutPickling):
-    __slots__ = []
+    __slots__: List[str] = []
 
     def __getstate__(self):
         return dict(
@@ -417,7 +427,7 @@ class DependentDictionary(object):
 
 # {{{ assertive accessors
 
-def one(iterable):
+def one(iterable: Iterable[T]) -> T:
     """Return the first entry of *iterable*. Assert that *iterable* has only
     that one entry.
     """
@@ -439,7 +449,10 @@ def one(iterable):
     return v
 
 
-def is_single_valued(iterable, equality_pred=operator.eq):
+def is_single_valued(
+        iterable: Iterable[T],
+        equality_pred: Callable[[T, T], bool] = operator.eq
+        ) -> bool:
     it = iter(iterable)
     try:
         first_item = next(it)
@@ -460,7 +473,10 @@ def all_roughly_equal(iterable, threshold):
             equality_pred=lambda a, b: abs(a-b) < threshold)
 
 
-def single_valued(iterable, equality_pred=operator.eq):
+def single_valued(
+        iterable: Iterable[T],
+        equality_pred: Callable[[T, T], bool] = operator.eq
+        ) -> T:
     """Return the first entry of *iterable*; Assert that other entries
     are the same with the first entry of *iterable*.
     """
@@ -484,7 +500,7 @@ def single_valued(iterable, equality_pred=operator.eq):
 
 # {{{ memoization / attribute storage
 
-def memoize(*args, **kwargs):
+def memoize(*args: F, **kwargs: Any) -> F:
     """Stores previously computed function values in a cache.
 
     Two keyword-only arguments are supported:
@@ -498,8 +514,10 @@ def memoize(*args, **kwargs):
 
     use_kw = bool(kwargs.pop('use_kwargs', False))
 
+    default_key_func: Optional[Callable[..., Any]]
+
     if use_kw:
-        def default_key_func(*inner_args, **inner_kwargs):
+        def default_key_func(*inner_args, **inner_kwargs):  # noqa pylint:disable=function-redefined
             return inner_args, frozenset(six.iteritems(inner_kwargs))
     else:
         default_key_func = None
@@ -544,6 +562,7 @@ def memoize(*args, **kwargs):
                 result = func(*args)
                 func._memoize_dic[args] = result  # pylint: disable=protected-access
                 return result
+
     if not args:
         return _deco
     if callable(args[0]) and len(args) == 1:
@@ -561,12 +580,9 @@ class _HasKwargs(object):
 
 def memoize_on_first_arg(function, cache_dict_name=None):
     """Like :func:`memoize_method`, but for functions that take the object
-    to do memoization as first argument.
+    in which do memoization information is stored as first argument.
 
     Supports cache deletion via ``function_name.clear_cache(self)``.
-
-    .. note::
-        *clear_cache* support requires Python 2.5 or newer.
     """
 
     if cache_dict_name is None:
@@ -600,22 +616,87 @@ def memoize_on_first_arg(function, cache_dict_name=None):
     return new_wrapper
 
 
-def memoize_method(method):
+def memoize_method(method: F) -> F:
     """Supports cache deletion via ``method_name.clear_cache(self)``.
-
-    .. note::
-        *clear_cache* support requires Python 2.5 or newer.
     """
 
     return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
 
 
+class keyed_memoize_on_first_arg:  # noqa: N801
+    """Like :func:`memoize_method`, but for functions that take the object
+    in which memoization information is stored as first argument.
+
+    Supports cache deletion via ``function_name.clear_cache(self)``.
+
+    :arg key: A function receiving the same arguments as the decorated function
+        which computes and returns the cache key.
+
+    .. versionadded :: 2020.3
+    """
+
+    def __init__(self, key, cache_dict_name=None):
+        self.key = key
+        self.cache_dict_name = cache_dict_name
+
+    def _default_cache_dict_name(self, function):
+        return intern("_memoize_dic_"
+                + function.__module__ + function.__name__)
+
+    def __call__(self, function):
+        cache_dict_name = self.cache_dict_name
+        key = self.key
+
+        if cache_dict_name is None:
+            cache_dict_name = self._default_cache_dict_name(function)
+
+        def wrapper(obj, *args, **kwargs):
+            cache_key = key(*args, **kwargs)
+
+            try:
+                return getattr(obj, cache_dict_name)[cache_key]
+            except AttributeError:
+                result = function(obj, *args, **kwargs)
+                setattr(obj, cache_dict_name, {cache_key: result})
+                return result
+            except KeyError:
+                result = function(obj, *args, **kwargs)
+                getattr(obj, cache_dict_name)[cache_key] = result
+                return result
+
+        def clear_cache(obj):
+            delattr(obj, cache_dict_name)
+
+        from functools import update_wrapper
+        new_wrapper = update_wrapper(wrapper, function)
+        new_wrapper.clear_cache = clear_cache
+
+        return new_wrapper
+
+
+class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
+    """Supports cache deletion via ``method_name.clear_cache(self)``.
+
+    :arg key: A function receiving the same arguments as the decorated function
+        which computes and returns the cache key.
+
+    .. versionadded :: 2020.3
+    """
+    def _default_cache_dict_name(self, function):
+        return intern("_memoize_dic_" + function.__name__)
+
+
 def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     """Supports cache deletion via ``method_name.clear_cache(self)``.
 
     :arg uncached_args: a list of argument numbers
         (0-based, not counting 'self' argument)
     """
+    from warnings import warn
+    warn("memoize_method_with_uncached is deprecated and will go away in 2022. "
+            "Use memoize_method_with_key instead",
+            DeprecationWarning,
+            stacklevel=2)
 
     if uncached_args is None:
         uncached_args = []
@@ -682,8 +763,9 @@ def memoize_method_nested(inner):
     """
 
     from warnings import warn
-    warn("memoize_method_nested is deprecated. Use @memoize_in(self, 'identifier') "
-            "instead", DeprecationWarning, stacklevel=2)
+    warn("memoize_method_nested is deprecated and will go away in 2021. "
+            "Use @memoize_in(self, 'identifier') instead", DeprecationWarning,
+            stacklevel=2)
 
     from functools import wraps
     cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
@@ -716,18 +798,22 @@ def memoize_method_nested(inner):
 
 class memoize_in(object):  # noqa
     """Adds a cache to a function nested inside a method. The cache is attached
-    to *object*.
+    to *container*.
 
-    Requires Python 2.5 or newer.
+    .. versionchanged :: 2020.3
+
+        *identifier* no longer needs to be a :class:`str`,
+        but it needs to be hashable.
     """
 
     def __init__(self, container, identifier):
-        key = "_pytools_memoize_in_dict_for_"+identifier
         try:
-            self.cache_dict = getattr(container, key)
+            memoize_in_dict = container._pytools_memoize_in_dict
         except AttributeError:
-            self.cache_dict = {}
-            setattr(container, key, self.cache_dict)
+            memoize_in_dict = {}
+            container._pytools_memoize_in_dict = memoize_in_dict
+
+        self.cache_dict = memoize_in_dict.setdefault(identifier, {})
 
     def __call__(self, inner):
         from functools import wraps
@@ -1152,6 +1238,10 @@ def wandering_element(length, wanderer=1, landscape=0):
 
 
 def indices_in_shape(shape):
+    from warnings import warn
+    warn("indices_in_shape is deprecated. You should prefer numpy.ndindex.",
+            DeprecationWarning, stacklevel=2)
+
     if isinstance(shape, int):
         shape = (shape,)
 
@@ -1305,17 +1395,17 @@ def enumerate_basic_directions(dimensions):
 # {{{ index mangling
 
 def get_read_from_map_from_permutation(original, permuted):
-    """With a permutation given by C{original} and C{permuted},
-    generate a list C{rfm} of indices such that
-    C{permuted[i] == original[rfm[i]]}.
+    """With a permutation given by *original* and *permuted*,
+    generate a list *rfm* of indices such that
+    ``permuted[i] == original[rfm[i]]``.
 
     Requires that the permutation can be inferred from
-    C{original} and C{permuted}.
+    *original* and *permuted*.
 
     .. doctest ::
 
-        >>> for p1 in generate_permutations(range(5)):
-        ...     for p2 in generate_permutations(range(5)):
+        >>> for p1 in generate_permutations(list(range(5))):
+        ...     for p2 in generate_permutations(list(range(5))):
         ...         rfm = get_read_from_map_from_permutation(p1, p2)
         ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
         ...         assert p2 == p2a
@@ -1332,17 +1422,17 @@ def get_read_from_map_from_permutation(original, permuted):
 
 
 def get_write_to_map_from_permutation(original, permuted):
-    """With a permutation given by C{original} and C{permuted},
-    generate a list C{wtm} of indices such that
-    C{permuted[wtm[i]] == original[i]}.
+    """With a permutation given by *original* and *permuted*,
+    generate a list *wtm* of indices such that
+    ``permuted[wtm[i]] == original[i]``.
 
     Requires that the permutation can be inferred from
-    C{original} and C{permuted}.
+    *original* and *permuted*.
 
-    .. doctest::
+    .. doctest ::
 
-        >>> for p1 in generate_permutations(range(5)):
-        ...     for p2 in generate_permutations(range(5)):
+        >>> for p1 in generate_permutations(list(range(5))):
+        ...     for p2 in generate_permutations(list(range(5))):
         ...         wtm = get_write_to_map_from_permutation(p1, p2)
         ...         p2a = [0] * len(p2)
         ...         for i, oi in enumerate(p1):
@@ -1831,7 +1921,8 @@ def generate_unique_names(prefix):
         try_num += 1
 
 
-def generate_numbered_unique_names(prefix, num=None):
+def generate_numbered_unique_names(
+        prefix: str, num: Optional[int] = None) -> Iterable[Tuple[int, str]]:
     if num is None:
         yield (0, prefix)
         num = 0
@@ -1853,18 +1944,20 @@ class UniqueNameGenerator(object):
     .. automethod:: add_names
     .. automethod:: __call__
     """
-    def __init__(self, existing_names=None, forced_prefix=""):
+    def __init__(self,
+            existing_names: Optional[Set[str]] = None,
+            forced_prefix: str = ""):
         if existing_names is None:
             existing_names = set()
 
         self.existing_names = existing_names.copy()
         self.forced_prefix = forced_prefix
-        self.prefix_to_counter = {}
+        self.prefix_to_counter: Dict[str, int] = {}
 
-    def is_name_conflicting(self, name):
+    def is_name_conflicting(self, name: str) -> bool:
         return name in self.existing_names
 
-    def _name_added(self, name):
+    def _name_added(self, name: str) -> None:
         """Callback to alert subclasses when a name has been added.
 
         .. note::
@@ -1874,7 +1967,7 @@ class UniqueNameGenerator(object):
         """
         pass
 
-    def add_name(self, name):
+    def add_name(self, name: str) -> None:
         if self.is_name_conflicting(name):
             raise ValueError("name '%s' conflicts with existing names")
         if not name.startswith(self.forced_prefix):
@@ -1883,11 +1976,11 @@ class UniqueNameGenerator(object):
         self.existing_names.add(name)
         self._name_added(name)
 
-    def add_names(self, names):
+    def add_names(self, names: Iterable[str]) -> None:
         for name in names:
             self.add_name(name)
 
-    def __call__(self, based_on="id"):
+    def __call__(self, based_on: str = "id") -> str:
         based_on = self.forced_prefix + based_on
 
         counter = self.prefix_to_counter.get(based_on, None)
